from typing import Awaitable, Optional, Sequence

import trio
from async_generator import asynccontextmanager
from serial.tools.list_ports import comports

from openrover.openrover_data import OpenRoverFirmwareVersion
from openrover.serial_trio import SerialTrio
from openrover.util import RoverDeviceNotFound
from .openrover_protocol import CommandVerb, OpenRoverProtocol
from .util import OpenRoverException

"""Communication settings for connecting to OpenRover hardware"""
DEFAULT_SERIAL_KWARGS = dict(baudrate=57600, stopbits=1)


def get_ftdi_device_paths() -> Sequence[str]:
    return [comport.device for comport in comports() if comport.manufacturer == "FTDI"]


async def get_openrover_protocol_version(
    device: SerialTrio,
) -> Awaitable[OpenRoverFirmwareVersion]:
    try:
        with trio.fail_after(1):
            orp = OpenRoverProtocol(device)
            while True:
                orp.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
                k, version = await orp.read_one()
                if k == 40:
                    return version
    except trio.TooSlowError as e:
        raise OpenRoverException(
            "Device did not respond to a request for version. Is it on?"
        ) from e
    except Exception as e:
        raise OpenRoverException("Device did not return a valid openrover version", e) from e


@asynccontextmanager
async def open_rover_device(*ports_to_try: Optional[str]):
    """
    Enumerates serial devices until it finds one that responds to a request for OpenRover version. Returns that device.
    :param ports_to_try: if provided, the devices to attempt to open (e.g. 'COM3'). Otherwise, all FTDI devices will be attempted
    :return: A SerialTrio device to use as a rover. If no appropriate device is found, will raise a RoverDeviceNotFound exception
    """
    exc_args = []
    for port in ports_to_try or get_ftdi_device_paths():
        async with SerialTrio(port, **DEFAULT_SERIAL_KWARGS) as device:
            try:
                await get_openrover_protocol_version(device)
                yield device
                return
            except OpenRoverException as e:
                exc_args.append((port, e))
    raise RoverDeviceNotFound(exc_args)



import enum
from typing import Any, Tuple

import trio

from .openrover_data import MOTOR_EFFORT_FORMAT, OPENROVER_DATA_ELEMENTS
from .serial_trio import SerialTrio
from .util import OpenRoverException

SERIAL_START_BYTE = bytes.fromhex("fd")


class CommandVerb(enum.IntEnum):
    NOP = 0
    GET_DATA = 10
    SET_FAN_SPEED = 20
    RESTART = 230
    CLEAR_SYSTEM_FAULT = 232
    SET_DRIVE_MODE = 240
    FLIPPER_CALIBRATE = 250
    RELOAD_SETTINGS = 1
    COMMIT_SETTINGS = 2
    SET_POWER_POLLING_INTERVAL_MS = 3
    SET_OVERCURRENT_THRESHOLD_100MA = 4
    SET_OVERCURRENT_TRIGGER_DURATION_5MS = 5
    SET_OVERCURRENT_RECOVERY_THRESHOLD_100MA = 6
    SET_OVERCURRENT_RECOVERY_DURATION_5MS = 7
    SET_PWM_FREQUENCY_KHZ = 8
    SET_BRAKE_ON_ZERO_SPEED_COMMAND = 9
    SET_BRAKE_ON_DRIVE_TIMEOUT = 11
    SET_MOTOR_SLOW_DECAY_MODE = 12
    SET_TIME_TO_FULL_SPEED_DECISECONDS = 13
    SET_PWM_FREQUENCY_100HZ = 14
    SET_SPEED_LIMIT_PERCENT = 15
    SET_OVERSPEED_ENCODER_THRESHOLD_ENCODER_100HZ = 16
    SET_OVERSPEED_DURATION_100MS = 17
    SET_BRAKE_ON_FAULT = 18


def encode_packet(*args: bytes):
    payload = b"".join(args)
    return SERIAL_START_BYTE + payload + bytes([checksum(payload)])


def checksum(values):
    return 255 - sum(values) % 255


class OpenRoverProtocol:
    def __init__(self, serial: SerialTrio):
        """Low-level communication for OpenRover"""
        self._serial = serial
        # A packet involves multiple read operations, so we must lock the device for reading
        self._read_lock = trio.StrictFIFOLock()

    async def read_one(self) -> Tuple[int, Any]:
        raw_data = await self._read_one_raw()
        data_element_index = raw_data[0]
        element_descriptor = OPENROVER_DATA_ELEMENTS[data_element_index]
        data_element_value = element_descriptor.data_format.unpack(raw_data[1:])
        return data_element_index, data_element_value

    async def _read_one_raw(self) -> bytes:
        """Reads a packet, verifies its checksum, and returns the packet payload"""
        async with self._read_lock:
            _ = await self._serial.read_until(SERIAL_START_BYTE)

            payload = await self._serial.read_exactly(3)

            (actual_checksum,) = await self._serial.read_exactly(1)

            expected_checksum = checksum(payload)
            if actual_checksum == expected_checksum:
                return payload
            else:
                raise OpenRoverException(
                    "Bad checksum {}, expected {}. Discarding data {}".format(
                        list(actual_checksum), list(expected_checksum), list(payload)
                    )
                )

    async def flush(self):
        await self._serial.flush(0)

    def write_nowait(
        self,
        motor_left: float,
        motor_right: float,
        flipper: float,
        command_verb: CommandVerb,
        command_arg: int,
    ):
        binary = encode_packet(
            MOTOR_EFFORT_FORMAT.pack(motor_left),
            MOTOR_EFFORT_FORMAT.pack(motor_right),
            MOTOR_EFFORT_FORMAT.pack(flipper),
            bytes([command_verb, command_arg]),
        )
        self._serial.write_nowait(binary)



from typing import Iterable, Tuple


class OpenRoverException(Exception):
    pass


class RoverDeviceNotFound(OpenRoverException):
    def __init__(self, devices_and_failures: Iterable[Tuple[str, Exception]]):
        self.devices_and_failures = devices_and_failures
        super().__init__()



import abc
import enum
import functools
import re
from typing import NamedTuple, Optional


class ReadDataFormat(abc.ABC):
    python_type = None

    @abc.abstractmethod
    def description(self):
        raise NotImplementedError

    @abc.abstractmethod
    def unpack(self, b: bytes):
        raise NotImplementedError


class WriteDataFormat(abc.ABC):
    python_type = None

    @abc.abstractmethod
    def description(self):
        raise NotImplementedError

    @abc.abstractmethod
    def pack(self, value) -> bytes:
        raise NotImplementedError


class IntDataFormat(ReadDataFormat, WriteDataFormat):
    def __init__(self, nbytes, signed):
        self.nbytes = nbytes
        self.signed = signed

    def description(self):
        s = "signed" if self.signed else "unsigned"
        n = self.nbytes * 8
        return f"{s} integer ({n} bits)"

    def pack(self, value):
        return int(value).to_bytes(self.nbytes, byteorder="big", signed=self.signed)

    def unpack(self, b: bytes):
        return int.from_bytes(b, byteorder="big", signed=self.signed)


OPENROVER_LEGACY_VERSION = 40621


@functools.total_ordering
class OpenRoverFirmwareVersion(NamedTuple):
    @classmethod
    def parse(cls, a_str):
        ver_re = re.compile(r"(\d+(?:[.]\d+){0,2})(?:-([^+])+)?(?:[+](.+))?", re.VERBOSE)
        match = ver_re.fullmatch(a_str)
        if match is None:
            raise ValueError
        parts = [int(p) for p in match.group(0).split(".")]
        return OpenRoverFirmwareVersion(*parts)

    major: int
    minor: int = 0
    patch: int = 0

    build: str = ""
    prerelease: str = ""

    @property
    def value(self):
        return self.major * 10000 + self.minor * 100 + self.patch * 10

    def __lt__(self, other):
        return (self.major, self.minor, self.patch, other.prerelease) < (
            other.major,
            other.minor,
            other.patch,
            self.prerelease,
        )

    def __str__(self):
        return (
            f"{self.major}.{self.minor}.{self.patch}"
            + (("-" + self.prerelease) if self.prerelease else "")
            + (("+" + self.build) if self.build else "")
        )


class DataFormatFirmwareVersion(ReadDataFormat):
    python_type = OpenRoverFirmwareVersion

    def unpack(self, b):
        v = UINT16.unpack(b)
        if v == OPENROVER_LEGACY_VERSION:
            return OpenRoverFirmwareVersion(1, 0, 0)
        return OpenRoverFirmwareVersion(v // 10000, v // 100 % 100, v % 10)

    def description(self):
        return (
            "XYYZZ, where X=major version, Y=minor version, Z = patch version."
            "e.g. 10502 = version 1.05.02. The special value 16421 represents pre-1.3 versions"
        )


class DataFormatChargerState(ReadDataFormat, WriteDataFormat):
    CHARGER_ACTIVE_MAGIC_BYTES = bytes.fromhex("dada")
    CHARGER_INACTIVE_MAGIC_BYTES = bytes.fromhex("0000")
    python_type = bool

    def pack(self, value):
        if value:
            return self.CHARGER_ACTIVE_MAGIC_BYTES
        else:
            return self.CHARGER_INACTIVE_MAGIC_BYTES

    def unpack(self, b):
        return bytes(b) == self.CHARGER_ACTIVE_MAGIC_BYTES

    def description(self):
        return "0xDADA if charging, else 0x0000"


class BatteryStatus(enum.Flag):
    overcharged_alarm = enum.auto()
    terminate_charge_alarm = enum.auto()
    over_temp_alarm = enum.auto()
    terminate_discharge_alarm = enum.auto()
    remaining_capacity_alarm = enum.auto()
    remaining_time_alarm = enum.auto()
    initialized = enum.auto()
    discharging = enum.auto()
    fully_charged = enum.auto()
    fully_discharged = enum.auto()


class DataFormatBatteryStatus(ReadDataFormat):
    python_type = BatteryStatus

    def unpack(self, b: bytes):
        assert len(b) == 2
        as_int = int.from_bytes(b, byteorder="big", signed=False)
        result = BatteryStatus(0)
        for mask, val in (
            (0x8000, BatteryStatus.overcharged_alarm),
            (0x4000, BatteryStatus.terminate_charge_alarm),
            (0x1000, BatteryStatus.over_temp_alarm),
            (0x0800, BatteryStatus.terminate_discharge_alarm),
            (0x0200, BatteryStatus.remaining_capacity_alarm),
            (0x0100, BatteryStatus.remaining_time_alarm),
            (0x0080, BatteryStatus.initialized),
            (0x0040, BatteryStatus.discharging),
            (0x0020, BatteryStatus.fully_charged),
            (0x0010, BatteryStatus.fully_discharged),
        ):
            if as_int & mask:
                result |= val
        return result

    def description(self):
        return "bit flags"


class DriveMode(enum.IntEnum):
    OPEN_LOOP = 0
    CLOSED_LOOP = 1


UINT16 = IntDataFormat(2, False)
INT16 = IntDataFormat(2, True)
UINT8 = IntDataFormat(1, signed=False)


class DataFormatFixedPrecision(ReadDataFormat, WriteDataFormat):
    """A fractional number packed as an integer, but representing a fractional number"""

    def __init__(self, base_type, step=1.0, zero=0.0):
        self.base_type = base_type
        # a change of 1 in the python type corresponds to a change of this many in the base type
        self.step = step
        # the value of 0 in the python type corresponds to this value in the base type
        self.zero = zero

    def unpack(self, b: bytes):
        n = self.base_type.unpack(b)
        return (n - self.zero) / self.step

    def pack(self, p):
        n = round(p * self.step + self.zero)
        return self.base_type.pack(n)

    def description(self):
        return "fractional (resolution=1/{}, zero={}) stored as {}".format(
            self.step, self.zero, self.base_type.description()
        )


class DataFormatDriveMode(ReadDataFormat):
    python_type = DriveMode

    def unpack(self, b: bytes):
        return DriveMode(UINT16.unpack(b))

    def pack(self, p: DriveMode):
        return UINT16.pack(p.value)

    def description(self):
        return DriveMode.__doc__


OLD_CURRENT_FORMAT = DataFormatFixedPrecision(UINT16, 34)

SIGNED_MILLIS_FORMAT = DataFormatFixedPrecision(INT16, 1000)
UNSIGNED_MILLIS_FORMAT = DataFormatFixedPrecision(UINT16, 1000)
OLD_VOLTAGE_FORMAT = DataFormatFixedPrecision(UINT16, 58)
FAN_SPEED_RESPONSE_FORMAT = DataFormatFixedPrecision(UINT16, 240)
DECIKELVIN_FORMAT = DataFormatFixedPrecision(UINT16, 10, zero=2731.5)
PERCENTAGE_FORMAT = DataFormatFixedPrecision(UINT16, 100)
MOTOR_EFFORT_FORMAT = DataFormatFixedPrecision(UINT8, 125, 125)
CHARGER_STATE_FORMAT = DataFormatChargerState()
FIRMWARE_VERSION_FORMAT = DataFormatFirmwareVersion()
DRIVE_MODE_FORMAT = DataFormatDriveMode()
BATTERY_STATUS_FORMAT = DataFormatBatteryStatus()


class MotorStatusFlag(enum.Flag):
    NONE = 0
    FAULT1 = enum.auto()
    FAULT2 = enum.auto()
    DECAY_MODE = enum.auto()
    REVERSE = enum.auto()
    BRAKE = enum.auto()
    COAST = enum.auto()


class DataFormatMotorStatus(ReadDataFormat):
    def description(self):
        return "motor status bit flags"

    def unpack(self, b: bytes):
        u = UINT16.unpack(b)

        bit_meanings = [
            MotorStatusFlag.FAULT1,
            MotorStatusFlag.FAULT2,
            MotorStatusFlag.DECAY_MODE,
            MotorStatusFlag.REVERSE,
            MotorStatusFlag.BRAKE,
            MotorStatusFlag.COAST,
        ]
        if len(bit_meanings) <= u.bit_length():
            raise ValueError("too many bits to unpack")

        result = MotorStatusFlag.NONE
        for i, flag in enumerate(bit_meanings):
            if u & 1 << i:
                result |= flag
        return result


class DataFormatIgnored(WriteDataFormat):
    def description(self):
        return f"Ignored data {self.n_bytes} bytes long"

    def pack(self, value=None) -> bytes:
        assert value is None
        return bytes(self.n_bytes)

    def __init__(self, n_bytes):
        self.n_bytes = n_bytes


class SystemFaultFlag(enum.Flag):
    NONE = 0
    OVERSPEED = enum.auto()
    OVERCURRENT = enum.auto()


class DataFormatSystemFault(ReadDataFormat):
    def description(self):
        return "System fault bit flags"

    def unpack(self, b: bytes):
        u = UINT16.unpack(b)

        bit_meanings = [SystemFaultFlag.OVERSPEED, SystemFaultFlag.OVERCURRENT]
        if len(bit_meanings) <= u.bit_length():
            raise ValueError("too many bits to unpack")

        result = SystemFaultFlag.NONE
        for i, flag in enumerate(bit_meanings):
            if u & 1 << i:
                result |= flag
        return result


class DataElement:
    def __init__(
        self,
        index: int,
        data_format: ReadDataFormat,
        name: str,
        description: str = None,
        not_implemented: bool = False,
        since: Optional[str] = None,
        until: Optional[str] = None,
    ):
        self.index = index
        self.data_format = data_format
        self.name = name
        self.description = description
        self.not_implemented = not_implemented
        self.since_version = None if since is None else OpenRoverFirmwareVersion.parse(since)
        self.until_version = None if until is None else OpenRoverFirmwareVersion.parse(until)

    def supported(self, version):
        if isinstance(version, str):
            v = OpenRoverFirmwareVersion.parse(version)
        elif isinstance(version, OpenRoverFirmwareVersion):
            v = version
        else:
            raise TypeError(
                f"Expected string or OpenRoverFirmwareVersion, but got {type(version)}"
            )

        if self.not_implemented:
            return False
        if self.since_version is not None and v < self.since_version:
            return False
        if self.until_version is not None:
            if self.until_version <= v:
                return False
        return True


elements = [
    DataElement(
        0, OLD_CURRENT_FORMAT, "battery (A+B) current (external)", "total current from batteries"
    ),
    DataElement(2, UINT16, "left motor speed", not_implemented=True),
    DataElement(4, UINT16, "right motor speed", not_implemented=True),
    DataElement(
        6,
        UINT16,
        "flipper position 1",
        "flipper position sensor 1. 0=15 degrees; 1024=330 degrees;",
    ),
    DataElement(
        8,
        UINT16,
        "flipper position 2",
        "flipper position sensor 2. 0=15 degrees; 1024=330 degrees;",
    ),
    DataElement(10, OLD_CURRENT_FORMAT, "left motor current"),
    DataElement(12, OLD_CURRENT_FORMAT, "right motor current"),
    DataElement(
        14,
        UINT16,
        "left motor encoder count",
        "May overflow or underflow. Increments when motor driven forward, decrements backward",
        since="1.4",
    ),
    DataElement(
        16,
        UINT16,
        "right motor encoder count",
        "May overflow or underflow. Increments when motor driven forward, decrements backward",
        since="1.4",
    ),
    DataElement(18, UINT16, "motors fault flag", not_implemented=True),
    DataElement(20, UINT16, "left motor temperature"),
    DataElement(22, UINT16, "right motor temperature", not_implemented=True),
    DataElement(24, OLD_VOLTAGE_FORMAT, "battery A voltage (external)"),
    DataElement(26, OLD_VOLTAGE_FORMAT, "battery B voltage (external)"),
    DataElement(
        28,
        UINT16,
        "left motor encoder interval",
        "0 when motor stopped. Else proportional to motor period (inverse motor speed)",
    ),
    DataElement(
        30,
        UINT16,
        "right motor encoder interval",
        "0 when motor stopped. Else proportional to motor period (inverse motor speed)",
    ),
    DataElement(
        32,
        UINT16,
        "flipper motor encoder interval",
        "0 when motor stopped. Else proportional to motor period (inverse motor speed)",
        not_implemented=True,
    ),
    DataElement(
        34,
        PERCENTAGE_FORMAT,
        "battery A state of charge",
        "Proportional charge, 0.0=empty, 1.0=full",
    ),
    DataElement(
        36,
        PERCENTAGE_FORMAT,
        "battery B state of charge",
        "Proportional charge, 0.0=empty, 1.0=full",
    ),
    DataElement(38, CHARGER_STATE_FORMAT, "battery charging state"),
    DataElement(40, FIRMWARE_VERSION_FORMAT, "release version"),
    DataElement(42, OLD_CURRENT_FORMAT, "battery A current (external)"),
    DataElement(44, OLD_CURRENT_FORMAT, "battery B current (external)"),
    DataElement(46, UINT16, "motor flipper angle"),
    DataElement(48, FAN_SPEED_RESPONSE_FORMAT, "fan speed"),
    DataElement(50, DRIVE_MODE_FORMAT, "drive mode", until="1.7"),
    DataElement(52, BATTERY_STATUS_FORMAT, "battery A status", since="1.2"),
    DataElement(54, BATTERY_STATUS_FORMAT, "battery B status", since="1.2"),
    DataElement(56, UINT16, "battery A mode", since="1.2"),
    DataElement(58, UINT16, "battery B mode", since="1.2"),
    DataElement(60, DECIKELVIN_FORMAT, "battery A temperature (internal)", since="1.2"),
    DataElement(62, DECIKELVIN_FORMAT, "battery B temperature (internal)", since="1.2"),
    DataElement(64, UNSIGNED_MILLIS_FORMAT, "battery A voltage (internal)", since="1.2"),
    DataElement(66, UNSIGNED_MILLIS_FORMAT, "battery B voltage (internal)", since="1.2"),
    DataElement(
        68,
        SIGNED_MILLIS_FORMAT,
        "battery A current (internal)",
        ">0 = charging; <0 = discharging",
        since="1.2",
    ),
    DataElement(
        70,
        SIGNED_MILLIS_FORMAT,
        "battery B current (internal)",
        ">0 = charging; <0 = discharging",
        since="1.2",
    ),
    DataElement(72, DataFormatMotorStatus(), "left motor status", since="1.7"),
    DataElement(74, DataFormatMotorStatus(), "right motor status", since="1.7"),
    DataElement(76, DataFormatMotorStatus(), "flipper motor status", since="1.7"),
    DataElement(78, FAN_SPEED_RESPONSE_FORMAT, "fan 1 duty", since="1.9"),
    DataElement(80, FAN_SPEED_RESPONSE_FORMAT, "fan 2 duty", since="1.9"),
    DataElement(82, DataFormatSystemFault(), "system fault flags", since="1.10"),
]

OPENROVER_DATA_ELEMENTS = {e.index: e for e in elements}


def strike(s):
    return f"~~{s}~~"


def doc():
    lines = ["| # | Name | Data Type | Description |", "| - | ---- | --------- | ----------- |"]

    for de in elements:
        lines.append(
            "|"
            + "|".join(
                [
                    strike(de.index) if de.not_implemented else de.index,
                    de.name,
                    de.data_format.description(),
                    de.description,
                ]
            )
            + "|"
        )
    return "\n".join(lines)


if __name__ == "__main__":
    print(doc())


def fix_encoder_delta(delta):
    MAX_ENCODER = 2 ** 16
    delta %= MAX_ENCODER
    if delta < MAX_ENCODER / 2:
        return delta
    else:
        return delta - MAX_ENCODER



import shutil
import subprocess
from pathlib import Path
from subprocess import list2cmdline

import pytest
import trio

import openrover
from openrover.find_device import open_rover_device
from openrover.openrover_data import OpenRoverFirmwareVersion
from openrover.openrover_protocol import CommandVerb, OpenRoverProtocol
from openrover.util import RoverDeviceNotFound


@pytest.fixture
async def device():
    try:
        async with open_rover_device() as dev:
            yield dev
    except RoverDeviceNotFound:
        pytest.skip("No openrover device found")


@pytest.fixture
def powerboard_firmware_file():
    p = Path(openrover.__path__[0], "tests/resources/PowerBoard-1.5.0.hex")
    assert p.is_file()
    return p


@pytest.fixture
def booty_exe():
    maybe_booty = shutil.which("booty")
    if maybe_booty is None:
        pytest.skip(
            "Could not test bootloader. Booty executable does not exist or is not in the"
            " executable path."
        )
    p = Path(maybe_booty)
    assert p.is_file()
    return p


async def test_reboot(device):
    orp = OpenRoverProtocol(device)
    try:
        orp.write_nowait(0, 0, 0, CommandVerb.RESTART, 0)

        # check that the device did turn off
        orp.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
        with pytest.raises(trio.TooSlowError):
            with trio.fail_after(1):
                await orp.read_one()

        # check that the device comes back up and starts responding to request for version
        with trio.fail_after(30):
            while True:
                orp.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
                with trio.move_on_after(1):
                    k, v = await orp.read_one()
                    if k == 40:
                        return
    except BaseException:
        await trio.sleep(30)
        raise


@pytest.mark.bootload
async def test_bootloader(powerboard_firmware_file, booty_exe):
    async with open_rover_device() as device:
        orp = OpenRoverProtocol(device)
        orp.write_nowait(0, 0, 0, CommandVerb.RESTART, 0)
        port = device.port
        # flash rover firmware
        args = [
            str(booty_exe),
            "--port",
            str(port),
            "--baudrate",
            "57600",
            "--hexfile",
            str(powerboard_firmware_file.absolute()),
            "--erase",
            "--load",
            "--verify",
        ]

    print("running bootloader: " + list2cmdline(args))

    with trio.fail_after(60 * 15):
        async with await trio.open_process(
            args, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as booty:
            async with trio.open_nursery() as nursery:

                async def check_stdout():
                    line_generator = stream_to_lines(booty.stdout)
                    lines = []
                    try:
                        while True:
                            with trio.fail_after(30):
                                a_line = await line_generator.__anext__()
                            lines.append(a_line)
                    except trio.TooSlowError:
                        pytest.fail(f"booty became unresponsive after output {lines}")
                    except StopAsyncIteration:
                        pass
                    await trio.sleep(1)
                    assert any(["device verified" in a_line for a_line in lines])

                async def check_stderr():
                    error_output = await stream_to_string(booty.stderr)
                    if "device not responding" in error_output:
                        pytest.fail("Rover did not respond to booty. Does it have a bootloader?")
                    assert error_output.strip() == ""

                async def check_retcode():
                    assert await booty.wait() == 0, "Booty exited with nonzero return code"

                nursery.start_soon(check_stderr)
                nursery.start_soon(check_stdout)
                nursery.start_soon(check_retcode)

    with trio.fail_after(30):
        while True:
            with trio.move_on_after(1):
                orp.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
                k, version = await orp.read_one()
                assert k == 40
                assert isinstance(version, OpenRoverFirmwareVersion)
                assert (version.major, version.minor, version.patch) == (1, 5, 0)


async def stream_to_string(stream: trio.abc.ReceiveStream):
    buf = ""
    async with stream:
        while True:
            new_bytes = await stream.receive_some(1)
            if not new_bytes:
                break
            buf += new_bytes.decode()
    return buf


async def stream_to_lines(stream: trio.abc.ReceiveStream):
    buf = ""
    async with stream:
        while True:
            new_bytes = await stream.receive_some(1)
            if not new_bytes:
                break
            buf += new_bytes.decode()
            *some_lines, buf = buf.splitlines()
            for line in some_lines:
                yield line
        if buf != "":
            yield buf



# from enum import IntEnum
# import logging
# import struct
# from typing import Any, Tuple
#
# import trio
#
# from openrover.serial_trio import SerialTrio
#
# _START_OF_FRAME = 0xf7
# _END_OF_FRAME = 0x7f
# _ESC = 0xf6
# _ESC_XOR = 0x20
#
#
# class BootyProtocolError(RuntimeError):
#     pass
#
#
# class BootyCommand(IntEnum):
#     READ_PLATFORM = 0x00
#     READ_VERSION = 0x01
#     READ_ROW_LEN = 0x02
#     READ_PAGE_LEN = 0x03
#     READ_PROG_LEN = 0x04
#     READ_MAX_PROG_SIZE = 0x05
#     READ_APP_START_ADDRESS = 0x06
#     READ_BOOT_START_ADDRESS = 0x07
#     READ_ADDR = 0x20
#     READ_MAX = 0x21
#     ERASE_PAGE = 0x10
#     WRITE_ROW = 0x30
#     WRITE_MAX = 0x31
#     START_APP = 0x40
#
#
# class BootyDeviceMetadata:
#     platform: str
#     version: str
#     row_len: int
#     page_len: int
#     prog_len: int
#     max_prog_size: int
#     app_start_address: int
#     boot_start_address: int
#
#
# class BootyProtocol:
#     _process_packets_task = None
#
#     def __init__(self, serial: SerialTrio):
#         self._serial = serial
#         # responses from the robot
#         self._queue = trio.Queue()
#
#     # region Booty Protocol actions
#     async def get_device_metadata(self):
#         with trio.open_nursery() as nursery:
#             platform = self._cmd_request_data(BootyCommand.READ_PLATFORM)
#             version = self._cmd_request_data(BootyCommand.READ_VERSION)
#             row_len = self._cmd_request_data(BootyCommand.READ_ROW_LEN)
#             page_len = self._cmd_request_data(BootyCommand.READ_PAGE_LEN)
#             prog_len = self._cmd_request_data(BootyCommand.READ_PROG_LEN)
#             max_prog_size = self._cmd_request_data(BootyCommand.READ_MAX_PROG_SIZE)
#             app_start_address = self._cmd_request_data(BootyCommand.READ_APP_START_ADDRESS)
#             boot_start_address = self._cmd_request_data(BootyCommand.READ_BOOT_START_ADDRESS)
#             metadata = BootyDeviceMetadata(
#                 platform=await platform,
#                 version=await version,
#                 row_len=await row_len,
#                 page_len=await page_len,
#                 prog_len=await prog_len,
#                 max_prog_size=await max_prog_size,
#                 app_start_address=await app_start_address,
#                 boot_start_address=await boot_start_address,
#             )
#
#         logging.info(f'Platform metadata = {metadata}')
#         return metadata
#
#     def cmd_start_app(self):
#         self._cmd_write(BootyCommand.START_APP)
#
#     def cmd_write_row(self, addr, instructions):
#         self._cmd_write(BootyCommand.WRITE_ROW, addr, instructions)
#
#     def cmd_write_max(self, addr, instructions):
#         self._cmd_write(BootyCommand.WRITE_MAX, addr, instructions)
#
#     def cmd_erase_page(self, addr):
#         self._cmd_write(BootyCommand.ERASE_PAGE, addr)
#
#     async def cmd_read_addr(self, addr):
#         await self._cmd_write(BootyCommand.READ_ADDR, addr)
#         return await self._cmd_expect_response(BootyCommand.READ_ADDR)
#
#     async def cmd_read_max(self, addr):
#         await self._cmd_write(BootyCommand.READ_MAX, addr)
#         return await self._cmd_expect_response(BootyCommand.READ_MAX)
#
#     # endregion
#
#     def _pack_data(self, verb, *args):
#         payload = struct.pack('!H')
#         if verb == BootyCommand.ERASE_PAGE:
#             payload += struct.pack('!H', *args)
#         elif verb in [BootyCommand.WRITE_ROW, BootyCommand.WRITE_MAX]:
#             addr, data = args
#             payload += struct.pack('!H', addr)
#             payload += b''.join(struct.pack('!L', instr) for instr in data)
#         else:
#             # most request packets have no associated data
#             assert len(args) == 0
#
#         return payload
#
#     def _unpack_data(self, payload: bytes) -> Tuple[BootyCommand, Any]:
#         verb = BootyCommand(*struct.unpack('!H', payload[:2]))
#         data = payload[2:]
#
#         if verb in [BootyCommand.READ_PLATFORM, BootyCommand.READ_VERSION]:
#             # interpret data as a string
#             return (verb, data.rstrip(b'\0').decode('utf-8'))
#         elif verb in [BootyCommand.READ_ADDR, BootyCommand.READ_MAX]:
#             # interpret data as a list of instructions (uint32's)
#             return (verb, [x for x, in struct.iter_unpack('!L', data)])
#         else:
#             # interpret data as a uint32
#             return (verb, *struct.unpack('!L', data))
#
#     def _remove_esc_chars(self, raw_message: bytes):
#         if _START_OF_FRAME in raw_message:
#             raise BootyProtocolError('Raw message contains start of frame character', raw_message)
#         if _END_OF_FRAME in raw_message:
#             raise BootyProtocolError('Raw message contains start of frame character', raw_message)
#         if raw_message[-1] == _ESC:
#             raise BootyProtocolError('Raw message ends with escape character', raw_message)
#
#         message = bytearray()
#         escape_next = False
#         for c in raw_message:
#             if escape_next:
#                 message.append(c ^ _ESC_XOR)
#                 escape_next = False
#             elif c == _ESC:
#                 escape_next = True
#             else:
#                 message.append(c)
#         return bytes(message)
#
#     async def _cmd_request_data(self, verb: BootyCommand):
#         assert verb in [BootyCommand.READ_PLATFORM, BootyCommand.READ_VERSION, BootyCommand.READ_ROW_LEN,
#                         BootyCommand.READ_PAGE_LEN, BootyCommand.READ_PROG_LEN, BootyCommand.READ_MAX_PROG_SIZE,
#                         BootyCommand.READ_APP_START_ADDRESS, BootyCommand.READ_BOOT_START_ADDRESS]
#         await self._cmd_write(verb)
#
#         reply_verb, data = await self._queue.get()
#         if verb != reply_verb:
#             raise BootyProtocolError(f'Requested data type {verb} but received {reply_verb} in response', data)
#         return data
#
#     async def _cmd_expect_response(self, expect_verb):
#         reply_verb, data = await self._queue.get()
#         if expect_verb != reply_verb:
#             raise BootyProtocolError(f'Requested data type {expect_verb} but received {reply_verb} in response', data)
#         return data
#
#     async def _cmd_write(self, verb, *args):
#         self._writer.write(
#             bytes([_START_OF_FRAME, *self._make_tx_packet(verb, *args), _END_OF_FRAME])
#         )
#         await self._writer.drain()
#
#     def _make_tx_packet(self, verb, *args):
#         """Create a data packet, not including framing bytes"""
#         pk = bytes([verb] + self._pack_data(verb, *args))
#         pk = pk + self._fletcher16_checksum(pk)
#         pk = pk.replace(bytes([_ESC]), bytes([_ESC, _ESC_XOR ^ _ESC]))
#         pk = pk.replace(bytes([_START_OF_FRAME]), bytes([_ESC, _ESC_XOR ^ _START_OF_FRAME]))
#         pk = pk.replace(bytes([_END_OF_FRAME]), bytes([_ESC, _ESC_XOR ^ _END_OF_FRAME]))
#         return pk
#
#     def _fletcher16_checksum(self, data: bytes):
#         sum1 = 0
#         sum2 = 0
#         for b in data:
#             sum1 = (sum1 + b) % 256
#             sum2 = (sum2 + sum1) % 256
#         return bytes([sum1, sum2])
#
#     async def _receive_packet(self):
#         before_frame_contents = await self._stream_reader.readuntil(bytes([_START_OF_FRAME]))[:-1]
#         if len(before_frame_contents):
#             logging.warning(f'Discarding data before start of frame {before_frame_contents}')
#         raw_message = await self._stream_reader.readuntil(bytes([_END_OF_FRAME]))[:-1]
#         logging.debug(f'Processing packet {raw_message}')
#
#         message = self._remove_esc_chars(raw_message)
#         if len(message) < 3:
#             raise BootyProtocolError('Packet too short', raw_message)
#         payload = message[:-2]
#         checksum = message[-2:]
#         expected_checksum = self._fletcher16_checksum(payload)
#         if checksum != expected_checksum:
#             raise BootyProtocolError('Packet failed checksum check', raw_message)
#
#         return self._unpack_data(payload)
#
#     async def iter_messages(self):
#         while True:
#             try:
#                 msg = await self._receive_packet()
#                 yield msg
#             except BootyProtocolError as e:
#                 logging.warning(f'Skipping packet because of error {e}')



import pytest


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config):
    config.inicfg["log_cli"] = "true"
    config.inicfg["log_cli_level"] = "NOTSET"
    config.inicfg["log_cli_format"] = "%(asctime)s %(levelname)s %(message)s"



import pytest
import pytest_trio.plugin

BOOTLOAD_OPT = "--bootloadok"
MOTOR_OPT = "--motorok"
BURNIN_OPT = "--burninok"


def pytest_addoption(parser):
    parser.addoption(
        BOOTLOAD_OPT,
        action="store_true",
        default=False,
        help=(
            "enable tests that may erase the rover's firmware. Even if all tests run to completion"
            " and succeed, you will need to reinstall the rover's firmware afterwards."
        ),
    )
    parser.addoption(
        MOTOR_OPT,
        action="store_true",
        default=False,
        help=(
            "enable tests that may activate the motors. "
            "The wheels should be able to spin freely during this test, "
            "so the rover should be untethered or raised on a jackstand."
        ),
    )
    parser.addoption(
        BURNIN_OPT,
        action="store_true",
        default=False,
        help=(
            "enable burn-in tests. "
            "These tests take a looong time and the rover should be raised on a jackstand."
        ),
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "motor: this test uses the rover motors")
    config.addinivalue_line("markers", "bootload: this test may reprogram the rover")
    config.addinivalue_line("markers", "burnin: run the long burn-in test")


def pytest_fixture_setup(fixturedef, request):
    return pytest_trio.plugin.handle_fixture(fixturedef, request, force_trio_mode=True)


def pytest_collection_modifyitems(config, items):
    pytest_trio.plugin.automark(items)

    if not config.getoption(MOTOR_OPT):
        skip_motors = pytest.mark.skip(reason="Motor tests are disabled by default.")
        for item in items:
            if "motor" in item.keywords:
                item.add_marker(skip_motors)

    if not config.getoption(BOOTLOAD_OPT):
        skip_bootload = pytest.mark.skip(reason=f"Need {BOOTLOAD_OPT} option to run")
        for item in items:
            if "bootload" in item.keywords:
                item.add_marker(skip_bootload)

    if not config.getoption(BURNIN_OPT):
        skip_burnin = pytest.mark.skip(reason=f"Need {BURNIN_OPT} option to run")
        for item in items:
            if "burnin" in item.keywords:
                item.add_marker(skip_burnin)



import pytest

from openrover.find_device import *


def test_ftdi_device_paths():
    d = get_ftdi_device_paths()
    for name in d:
        assert isinstance(name, str)
        assert name != ""


async def test_open_any_openrover_device():
    if len(get_ftdi_device_paths()) == 0:
        pytest.skip("no FTDI devices found")

    async with open_rover_device() as device:
        assert isinstance(device, SerialTrio)
        # the device must still be open here
        assert await get_openrover_protocol_version(device) is not None

    # the device must be closed afterwards
    with pytest.raises(OpenRoverException):
        await get_openrover_protocol_version(device)


async def test_open_rover_device_sequentially_okay():
    if len(get_ftdi_device_paths()) == 0:
        pytest.skip("no FTDI devices found")

    for i in range(3):
        async with open_rover_device():
            pass


async def test_open_rover_device_nested_fails():
    if len(get_ftdi_device_paths()) == 0:
        pytest.skip("no FTDI devices found")

    async with open_rover_device() as d:
        with pytest.raises(OpenRoverException):
            async with SerialTrio(d.port):
                pass


async def test_missing_device():
    with pytest.raises(OpenRoverException):
        async with SerialTrio("/dev/nosuchdevice") as d2:
            pass



import errno
import logging
import warnings

import serial
import serial.tools
import serial.tools.list_ports
import trio

from .util import OpenRoverException


class OpenRoverWarning(RuntimeWarning):
    pass


class DeviceClosedException(serial.SerialException):
    pass


class SerialTrio(trio.abc.AsyncResource):
    _serial = None  # type: serial.Serial
    _inbound_high_water = 4000
    _outbound_high_water = 8000

    def __init__(self, port, **serial_kwargs):
        """Wrapper for pyserial that makes it work better with async"""
        self.port = port
        self.serial_kwargs = {
            "write_timeout": 0,
            "inter_byte_timeout": None,
            "timeout": 0,
            "exclusive": True,
        }
        self.serial_kwargs.update(serial_kwargs)

        try:
            self._serial = serial.Serial(self.port, **self.serial_kwargs)
        except serial.SerialException as e:
            if e.errno == errno.EAGAIN:
                raise OpenRoverException("Serial device is already open", self.port) from e
            if e.errno == errno.ENOENT:
                raise OpenRoverException(
                    "Could not connect to serial device - file not found. Is it connected?",
                    self.port,
                ) from e
            if e.errno == errno.EACCES:
                raise OpenRoverException(
                    "Access error when trying to connect to serial device. Is it open in another"
                    " process? Does this user have OS permission?",
                    self.port,
                ) from e
            if e.errno == errno.EISDIR or e.errno == errno.ENOTTY:
                raise OpenRoverException("Does not appear to be a serial device") from e
            raise OpenRoverException("Could not connect to serial device.") from e

    @property
    def in_waiting(self):
        try:
            return self._serial.in_waiting
        except Exception as e:
            if not self._serial.is_open:
                raise DeviceClosedException from e
            raise

    def _read_bytes_nowait(self, n_max):
        if self._inbound_high_water <= self.in_waiting:
            warnings.warn(
                "Incoming buffer is backlogged. Data may be lost. {} bytes".format(
                    self._serial.in_waiting
                )
            )
        return self._serial.read(n_max)

    async def read_until(self, terminator):
        terminator = bytes(terminator)
        assert terminator != b""
        line = bytearray()
        try:
            while not line.endswith(terminator):
                line.extend(self._read_bytes_nowait(1))
                await trio.sleep(0.001)
            return bytes(line)
        except trio.Cancelled:
            logging.exception(f"Abandoning data: {line}")
            self._serial.cancel_read()
            raise

    async def read_exactly(self, count):
        line = bytearray()
        while len(line) < count:
            line.extend(self._read_bytes_nowait(count - len(line)))
            await trio.sleep(0.001)
        return bytes(line)

    def write_nowait(self, data):
        self._serial.write(data)
        if self._outbound_high_water <= self._serial.out_waiting:
            warnings.warn(
                "Outgoing buffer is backlogged. Data may be lost. {} bytes".format(
                    self._serial.out_waiting
                )
            )

    async def write(self, data):
        self._serial.write(data)
        if self._outbound_high_water <= self._serial.out_waiting:
            warnings.warn(
                "Outgoing buffer is backlogged. Data may be lost. {} bytes".format(
                    self._serial.out_waiting
                )
            )
        try:
            await self.flush()
        except trio.Cancelled():
            self._serial.cancel_write()
            raise

    async def flush(self, n_bytes=0):
        """wait until the number of queued outgoing bytes is less than or equal to n_bytes"""
        assert n_bytes >= 0
        while self._serial.out_waiting > n_bytes:
            await trio.sleep(0.001)

    async def aclose(self):
        try:
            if self._serial.is_open:
                await self.flush()
        finally:
            self._serial.close()
            assert not self._serial.is_open



from typing import Any, Dict, Iterable, Optional

import trio
from async_generator import asynccontextmanager

from openrover.find_device import open_rover_device
from openrover.openrover_data import OPENROVER_DATA_ELEMENTS
from .openrover_protocol import CommandVerb, OpenRoverProtocol
from .serial_trio import SerialTrio
from .util import OpenRoverException


@asynccontextmanager
async def open_rover(path_to_serial: Optional[str] = None):
    args = [] if path_to_serial is None else [path_to_serial]

    async with open_rover_device(*args) as device:
        rover = Rover()
        await rover.set_device(device)
        yield rover


class Rover:
    _motor_left = 0
    _motor_right = 0
    _motor_flipper = 0

    _rover_protocol = None
    _device = None

    def __init__(self):
        """An OpenRover object"""
        self._motor_left = 0
        self._motor_right = 0
        self._motor_flipper = 0
        self._openrover_data_to_memory_channel = {
            i: trio.open_memory_channel(0) for i in OPENROVER_DATA_ELEMENTS.keys()
        }

    async def set_device(self, device: SerialTrio):
        self._device = device
        self._rover_protocol = OpenRoverProtocol(device)

    def set_motor_speeds(self, left, right, flipper):
        assert -1 <= left <= 1
        assert -1 <= right <= 1
        assert -1 <= flipper <= 1
        self._motor_left = left
        self._motor_right = right
        self._motor_flipper = flipper

    def _send_command(self, cmd, arg):
        self._rover_protocol.write_nowait(
            self._motor_left, self._motor_right, self._motor_flipper, cmd, arg
        )

    def send_speed(self):
        self._send_command(CommandVerb.NOP, 0)

    def set_fan_speed(self, fan_speed):
        assert 0 <= fan_speed <= 1
        self._send_command(CommandVerb.SET_FAN_SPEED, int(fan_speed * 240))

    def clear_system_fault(self):
        self._send_command(CommandVerb.CLEAR_SYSTEM_FAULT, 0)

    def flipper_calibrate(self):
        self._send_command(CommandVerb.FLIPPER_CALIBRATE, int(CommandVerb.FLIPPER_CALIBRATE))

    async def get_data(self, index) -> Any:
        """Get the next value for the given data index.
        The type of the returned value depends on the index passed."""
        self._send_command(CommandVerb.GET_DATA, index)
        with trio.fail_after(1):
            k, data = await self._rover_protocol.read_one()
            if k != index:
                raise OpenRoverException(
                    f"Received unexpected data. Expected {index}, received {k}:{data}"
                )

        return data

    async def get_data_items(self, indices: Iterable[int]) -> Dict[int, Any]:
        indices = sorted(set(indices))
        result = dict.fromkeys(indices)

        for index in indices:
            self._send_command(CommandVerb.GET_DATA, index)
        for index in indices:
            with trio.fail_after(1):
                k, data = await self._rover_protocol.read_one()
                if k != index:
                    raise OpenRoverException(
                        "Received unexpected data. Expected {}, received {}:{}".format(
                            index, k, data
                        )
                    )
            result[k] = data

        return result


async def get_openrover_version(port):
    try:
        with trio.fail_after(1):
            async with SerialTrio(port, baudrate=57600) as device:
                orp = OpenRoverProtocol(device)
                for i in range(2):
                    orp.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
                    k, version = await orp.read_one()
                    if k == 40:
                        return version
    except Exception as e:
        raise OpenRoverException("Did not respond to request for OpenRover version") from e



# suppress warnings and print output correctly on Ubuntu. Must be done
# before importing trio
# https://github.com/python-trio/trio/issues/1065
import sys

if sys.excepthook.__name__ == "apport_excepthook":
    sys.excepthook = sys.__excepthook__

from .openrover_protocol import OpenRoverProtocol
from .rover import open_rover, Rover
from .util import OpenRoverException

name = "openrover"

__all__ = ["OpenRoverException", "OpenRoverProtocol", "open_rover", "Rover"]



import argparse
import subprocess
import sys
from pathlib import Path

import trio

from openrover import OpenRoverProtocol
from openrover.find_device import get_ftdi_device_paths
from openrover.openrover_data import OpenRoverFirmwareVersion
from openrover.openrover_protocol import CommandVerb
from openrover.serial_trio import SerialTrio

BAUDRATE = 57600

SETTINGS_VERBS = list(map(CommandVerb, [*range(3, 10), *range(11, 19)]))


def rover_command_arg_pair(arg):
    k, v = arg.split(":", 2)
    k = CommandVerb(int(k))
    if k not in SETTINGS_VERBS:
        raise ValueError
    if not 0 <= int(v) <= 255:
        raise ValueError
    return k, int(v)


async def amain():
    parser = argparse.ArgumentParser(
        description="OpenRover companion utility to bootload robot and configure settings.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    pitstop_action = parser.add_subparsers(dest="action", required=True, metavar="action")
    flash = pitstop_action.add_parser(
        "flash", help="write the given firmware hex file onto the rover"
    )
    flash.add_argument(
        type=argparse.FileType("r"), dest="hexfile", help="*.hex file containing the new firmware",
    )

    checkversion = pitstop_action.add_parser(
        "checkversion", help="Check the version of firmware installed",
    )
    checkversion.add_argument(
        nargs="?",
        type=OpenRoverFirmwareVersion.parse,
        dest="version_expected",
        metavar="X | X.Y | X.Y.Z",
        help="Minimum acceptable version",
    )

    test = pitstop_action.add_parser(
        "test",
        help="Run tests on the rover",
        description=(
            "Run tests on the rover. Some tests are inherently unsafe and are disabled by default,"
            " but may be enabled with below flags"
        ),
    )
    test.add_argument(
        "--bootloadok",
        action="store_true",
        help="Allow tests that may alter the firmware version installed on the rover",
    )
    test.add_argument(
        "--motorok",
        action="store_true",
        help=(
            "Allow short-running tests that spin the rover motors. Note the rover should be up on"
            " blocks for the duration of this test so it does not drive away."
        ),
    )
    test.add_argument(
        "--burninok",
        action="store_true",
        help=(
            "Allow the long-running burn-in test to verify hardware reliability. The rover's"
            " wheels will spin during testing. Note the rover should be up on blocks for the"
            " duration of this test so it does not drive away."
        ),
    )

    config = pitstop_action.add_parser("config", help="Update rover persistent settings")
    config.add_argument(
        type=rover_command_arg_pair,
        dest="config_items",
        metavar="k:v",
        nargs="*",
        help="Send additional commands to the rover. v may be 0-255; k may be:\n\t"
        + "\n\t".join("{}={}".format(s.value, s.name) for s in SETTINGS_VERBS),
    )
    config.add_argument(
        "--commit",
        action="store_true",
        help=(
            "Persist these config options across reboot. By default, settings persist only until"
            " rover is restarted."
        ),
    )

    parser.add_argument(
        "-p",
        "--port",
        type=str,
        help="Which device to use. If omitted, we will search for a possible rover device",
        metavar="port",
    )
    args = parser.parse_args()

    port = args.port
    if port is None:
        print("Scanning for possible rover devices")
        ports = get_ftdi_device_paths()
        if len(ports) == 0:
            print("No devices found")
            sys.exit(1)
        if len(ports) > 1:
            print(f"Multiple devices found: {', '.join(ports)}")
        port = ports[0]
    print(f"Using device {port}")

    if args.action == "flash":
        async with SerialTrio(port, baudrate=BAUDRATE) as ser:
            orp = OpenRoverProtocol(ser)
            print("instructing rover to restart")
            for i in range(3):
                orp.write_nowait(0, 0, 0, CommandVerb.RESTART, 0)
            await orp.flush()

        pargs = [
            sys.executable,
            "-m",
            "booty",
            "--port",
            port,
            "--baudrate",
            str(BAUDRATE),
            "--hexfile",
            args.hexfile.name,
            "--erase",
            "--load",
            "--verify",
        ]
        print(f"invoking bootloader: {subprocess.list2cmdline(pargs)}")
        subprocess.check_call(pargs)

        print("starting firmware")
        async with SerialTrio(port, baudrate=BAUDRATE) as ser:
            ser.write_nowait(bytes.fromhex("f701004041437f"))
        print(
            "\n".join(
                [
                    r"      VROOM      ",
                    r"  _           _  ",
                    r" /#\ ------- /#\ ",
                    r" |#|  (o=o)  |#| ",
                    r" \#/ ------- \#/ ",
                    r"                 ",
                ]
            )
        )

    elif args.action == "checkversion":
        actual_version = None
        async with SerialTrio(port, baudrate=BAUDRATE) as device:
            orp = OpenRoverProtocol(device)
            orp.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
            with trio.move_on_after(10):
                k, version = await orp.read_one()
                if k == 40:
                    actual_version = version

        if actual_version is None:
            print("Could not get version of attached rover")
            sys.exit(1)
        else:
            print(f"Firmware version installed = {actual_version}")

        if args.version_expected is not None:
            print(f"Firmware version expected >= {args.version_expected}")
            if args.version_expected <= actual_version:
                print(f"Passed :-)")
            if actual_version < args.version_expected:
                print(f"Failed :-(")
                sys.exit(1)

    elif args.action == "config":
        async with SerialTrio(port, baudrate=57600) as device:
            orp = OpenRoverProtocol(device)
            print("Reloading settings from non-volatile memory.")
            orp.write_nowait(0, 0, 0, CommandVerb.RELOAD_SETTINGS, 0)
            for k, v in args.config_items or ():
                print(f"\tSetting {k.value} ({k.name}) = {v}")
                orp.write_nowait(0, 0, 0, k, v)
            if args.commit:
                print(
                    "Committing settings to non-volatile memories. "
                    "These new settings will persist on reboot."
                )
                orp.write_nowait(0, 0, 0, CommandVerb.COMMIT_SETTINGS, 0)
            else:
                print("These new settings will be reset on reboot.")
            await orp.flush()

    elif args.action == "test":
        argflags = []
        for argname in ("bootloadok", "burninok", "motorok"):
            if getattr(args, argname):
                argflags.append(f"--{argname}")

        from . import tests

        wd = Path(tests.__file__).parent.absolute()
        completed = await trio.run_process(
            [sys.executable, "-m", "pytest", *argflags], check=False, cwd=wd
        )
        sys.exit(completed.returncode)


def main():
    trio.run(amain)


if __name__ == "__main__":
    main()



import statistics

import pytest
import trio

from openrover.find_device import open_rover_device
from openrover.openrover_data import OpenRoverFirmwareVersion
from openrover.openrover_protocol import CommandVerb, OpenRoverProtocol
from openrover.util import RoverDeviceNotFound

n = 100


@pytest.fixture
async def protocol():
    try:
        async with open_rover_device() as r:
            yield OpenRoverProtocol(r)
    except RoverDeviceNotFound:
        pytest.skip("This test requires a rover device but none was found")


async def test_rtt(protocol):
    times = []
    for _ in range(n):
        protocol.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
        await protocol.flush()
        t0 = trio.current_time()
        await protocol.read_one()
        t1 = trio.current_time()
        times.append(t1 - t0)
    assert 0.010 < statistics.mean(times) < 0.030
    assert 0 < statistics.stdev(times) < 0.030


async def test_protocol_write_read_immediate(protocol):
    n_received = 0

    for i in range(n):
        protocol.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
        with trio.fail_after(1):
            key, version = await protocol.read_one()
            assert key == 40
            assert isinstance(version, OpenRoverFirmwareVersion)
            assert isinstance(version.value, int)
            assert 0 < version.value
            n_received += 1

    print(f"success ratio {(n_received / n)}")
    assert 0.9 < n_received / n <= 1


async def test_protocol_writes_then_reads(protocol):
    n_received = 0
    for _ in range(n):
        protocol.write_nowait(0, 0, 0, CommandVerb.GET_DATA, 40)
    try:
        for i in range(n):
            with trio.fail_after(5):
                key, version = await protocol.read_one()
                assert key == 40
                assert isinstance(version, OpenRoverFirmwareVersion)
                assert isinstance(version.value, int)
                assert 0 < version.value
                n_received += 1
    except trio.TooSlowError:
        pass

    print(f"success ratio {(n_received / n)}")
    assert 0.9 < n_received / n <= 1


async def test_responses_sequential(protocol):
    keys = [14, 16, 28, 30]

    for k in keys:
        protocol.write_nowait(0, 0, 0, CommandVerb.GET_DATA, k)

    result_keys = []
    for i in range(4):
        with trio.fail_after(1):
            k, _ = await protocol.read_one()
        result_keys.append(k)
    assert keys == result_keys



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openrover', 'openrover.tests', 'openrover.tests.burnin']

package_data = \
{'': ['*'], 'openrover.tests': ['resources/*']}

install_requires = \
['async_generator>=1.10,<2.0',
 'booty>=0.3.0,<0.4.0',
 'pyserial>=3.4,<4.0',
 'pytest-trio>=0.6.0,<0.7.0',
 'pytest>=5.4.3,<6.0.0',
 'trio>=0.16.0,<0.17.0']

entry_points = \
{'console_scripts': ['openrover-tests = openrover:main',
                     'pitstop = openrover.pitstop:main']}

setup_kwargs = {
    'name': 'openrover',
    'version': '1.0.0b1',
    'description': 'A Python driver for driving the Rover Robotics OpenRover Basic robot',
    'long_description': '# OpenRover Python Suite\n\nThis is the official Python driver for the [Rover Robotics](https://roverrobotics.com/) "Open Rover Basic" robot. Use this as a starting point to get up and running quickly.\n\nIncluded in this package are:\n\n1. A Python library for programmatically interfacing with the Rover over USB\n2. A command line application "`pitstop`" for upgrading and configuring the Rover firmware\n3. A test suite that confirms the Firmware and hardware are operating as expected.\n\n## Setup\n\nTo install official releases from PyPi:\n\n```shell script\npython3 -m pip install -U pip setuptools\npython3 -m pip install -U openrover --no-cache-dir\n```\n\nOn Linux, you may not have permission to access USB devices. If this is the case, run the following then restart your computer:\n\n```shell script\nsudo usermod -a -G dialout $(whoami)\n```\n\n### Development setup\n\nManual Prerequisites:\n\n* Python3 (recommended to install Python3.6, Python3.7, and Python3.8)\n* [Poetry](https://python-poetry.org/docs/#installation) \n\nInstead, we recommend:\n\n```\ngit clone https://github.com/RoverRobotics/openrover-python.git\ncd openrover-python\npoetry install\n```\n\n#### Useful commands\n\nFor development testing, it is recommended to use `tox`, which can run tests on multiple Python interpreters.\n\n<dl>\n    <dt><code>pytest</code></dt>\n    <dd>Test on current Python interpreter</dd>\n    <dt><code>tox</code></dt>\n    <dd>Test on all supported Python minor versions</dd>\n    <dt><code>black .</code></dt>\n    <dd>Reformat code to a uniform style</dd>\n    <dt><code>githooks setup</code></dt>\n    <dd>Install git pre-commit hook to automatically run <code>black</code></dd>\n</dl>\n\n### Caveats\n\n* When running in PyCharm in debug mode, you will get a warning like "RuntimeWarning: You seem to already have a custom sys.excepthook handler installed ..." https://github.com/python-trio/trio/issues/1553\n* Note this is a pyproject (PEP-517) project so it will NOT work to `pip install --editable ...` for development. Instead use `poetry install` as above.\n\n\n### pitstop\n\nPitstop is a helper program to bootload your rover and set options. After installing the openrover package, you can invoke it with `pitstop` or `python3 -m openrover.pitstop`.\n\n```text\n> pitstop --help\n  usage: pitstop [-h] [-p port] action ...\n  \n  OpenRover companion utility to bootload robot and configure settings.\n  \n  positional arguments:\n    action\n      flash               write the given firmware hex file onto the rover\n      checkversion        Check the version of firmware installed\n      test                Run tests on the rover\n      update              Update rover persistent settings\n  \n  optional arguments:\n    -h, --help            show this help message and exit\n    -p port, --port port  Which device to use. If omitted, we will search for a possible rover device\n```\n\n### tests\n\nTo run tests, first attach the rover via breakout cable then run `pitstop test`.\nBy default, tests that involve running the motors will be skipped, since you may not want a rover ripping cables out of your computer. If you have made sure running the motors will not damage anything, these tests can be opted in with the flag `--motorok`.\n\n```text\n> pitstop test --motorok\nScanning for possible rover devices\nUsing device /dev/ttyUSB0\n========================== test session starts ============================\nplatform linux -- Python 3.8.2, pytest-5.4.3, py-1.9.0, pluggy-0.13.1\nrootdir: /home/dan/Documents/openrover-python/openrover\nplugins: trio-0.6.0\ncollected 73 items                                                                                                                                                                                           \n\ntests/test_bootloader.py .s                                                                                                                                                                            [  2%]\ntests/test_find_device.py .....                                                                                                                                                                        [  9%]\ntests/test_openrover_protocol.py ....                                                                                                                                                                  [ 15%]\ntests/test_rover.py ..................x.x.........x................Xxxx..........                                                                                                                      [ 98%]\ntests/burnin/test_burnin.py s                                                                                                                                                                          [100%]\n\n===== 64 passed, 2 skipped, 6 xfailed, 1 xpassed in 83.94s (0:01:23) =====\n```\n\n![OpenRover Basic](https://docs.roverrobotics.com/1-manuals/0-cover-photos/1-open-rover-basic-getting-started-vga.jpg)\n',
    'author': 'Dan Rose',
    'author_email': 'dan@digilabs.io',
    'maintainer': 'Dan Rose',
    'maintainer_email': 'dan@digilabs.io',
    'url': 'https://github.com/RoverRobotics/openrover_python_driver',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import statistics

import pytest
import trio

from openrover.openrover_data import (
    fix_encoder_delta,
    MotorStatusFlag,
    OPENROVER_DATA_ELEMENTS,
    OpenRoverFirmwareVersion,
    SystemFaultFlag,
)
from openrover.rover import open_rover, Rover
from openrover.util import OpenRoverException, RoverDeviceNotFound


@pytest.fixture
async def rover():
    try:
        async with open_rover() as r:
            yield r
    except RoverDeviceNotFound:
        pytest.skip("This test requires a rover device but none was found")


async def test_find_openrover(rover):
    assert rover is not None
    assert isinstance(rover, Rover)


async def test_get_version(rover):
    version = await rover.get_data(40)
    assert isinstance(version, OpenRoverFirmwareVersion)

    version2 = await rover.get_data(40)
    assert version == version2
    assert 0 <= version.major <= 100
    assert 0 <= version.minor <= 100
    assert 0 <= version.patch <= 100


async def test_recover_from_bad_data(rover):
    await rover._rover_protocol._serial.write(b"test" * 20)

    for i in range(3):
        try:
            result = await rover.get_data(40)
            if result is not None:
                return
        except (OpenRoverException, trio.TooSlowError):
            pass
    assert False


async def test_missing_device():
    with pytest.raises(OpenRoverException):
        async with open_rover("missing_device"):
            pass


@pytest.mark.parametrize("fan_speed_cmd", [0.1, 0.4, 0.7])
async def test_fan_speed(rover, fan_speed_cmd):
    version = await rover.get_data(40)

    rover.set_fan_speed(fan_speed_cmd)
    assert await rover.get_data(48) == fan_speed_cmd
    if version < OpenRoverFirmwareVersion(1, 9):
        await trio.sleep(3)
        # manual fan duty should revert to zero after timeout
        assert await rover.get_data(48) == 0
    elif version < OpenRoverFirmwareVersion(1, 10):
        pytest.xfail("Fan speed control not implemented.")
    else:
        speeds_a = []
        speeds_b = []

        # wait a moment for fan duty to catch up with commanded value
        await trio.sleep(0.1)

        for i in range(int(fan_speed_cmd * 40)):
            speeds_a.append(await rover.get_data(78))
            speeds_b.append(await rover.get_data(80))
            await trio.sleep(0.5)

        # fan speeds should all be between 0 and 1
        assert all(0 <= x <= 1 for x in speeds_a)
        assert all(0 <= x <= 1 for x in speeds_b)

        # fan duty should start at the commanded value
        assert speeds_a[0] == pytest.approx(fan_speed_cmd, abs=0.05)
        assert speeds_b[0] == pytest.approx(fan_speed_cmd, abs=0.05)

        # fan duty should decrease
        assert speeds_a == sorted(speeds_a, reverse=True)
        assert speeds_b == sorted(speeds_b, reverse=True)

        # fan duty should end near zero
        assert speeds_a[-1] == pytest.approx(0, abs=0.05)
        assert speeds_b[-1] == pytest.approx(0, abs=0.05)


async def test_get_all_data_elements(rover):
    version = await rover.get_data(40)

    for i, de in OPENROVER_DATA_ELEMENTS.items():
        if de.supported(version):
            v = await rover.get_data(i)
            assert v is not None


@pytest.mark.motor
async def test_overspeed_fault(rover):
    v = await rover.get_data(40)
    if not OPENROVER_DATA_ELEMENTS[82].supported(v):
        pytest.xfail("System Fault Flag not implemented in this version")

    rover.clear_system_fault()
    assert await rover.get_data(82) == SystemFaultFlag.NONE

    enc_counts_left = []
    enc_counts_right = []
    enc_intervals_left = []
    enc_intervals_right = []
    rover.set_motor_speeds(1.0, 1.0, 0)
    for _ in range(50):
        rover.send_speed()
        await trio.sleep(0.1)
        data = await rover.get_data_items([14, 16, 28, 30])
        enc_counts_left.append(data[14])
        enc_counts_right.append(data[16])
        enc_intervals_left.append(data[28])
        enc_intervals_right.append(data[30])

    # motors running at full speed
    for x in enc_intervals_left[5:20]:
        assert x < 100
    for x in enc_intervals_right[5:20]:
        assert x < 100

    # motors should come to a stop
    for x in enc_intervals_left[40:]:
        assert x > 1000 or x == 0
    for x in enc_intervals_right[40:]:
        assert x > 1000 or x == 0

    assert await rover.get_data(82) == SystemFaultFlag.OVERSPEED

    enc_intervals_left2 = []
    enc_intervals_right2 = []
    rover.set_motor_speeds(0.2, 0.2, 0)
    for _ in range(10):
        rover.send_speed()
        await trio.sleep(0.1)
        data = await rover.get_data_items([28, 30])
        enc_intervals_left2.append(data[28])
        enc_intervals_right2.append(data[30])

    # fault condition still active so motor speeds should be ignored
    assert all(x == 0 or x > 1000 for x in enc_intervals_left2)
    assert all(x == 0 or x > 1000 for x in enc_intervals_right2)
    assert await rover.get_data(82) == SystemFaultFlag.OVERSPEED

    rover.clear_system_fault()
    assert await rover.get_data(82) == SystemFaultFlag.NONE


@pytest.mark.parametrize("motor_effort", [0, -0.1, +0.1, -0.2, +0.2, 0])
@pytest.mark.motor
async def test_encoder_intervals(rover, motor_effort):
    version = await rover.get_data(40)
    counts_supported = all(OPENROVER_DATA_ELEMENTS[i].supported(version) for i in (14, 16))

    enc_counts_left = []
    enc_counts_right = []
    enc_intervals_left = []
    enc_intervals_right = []

    rover.set_motor_speeds(motor_effort, motor_effort, 0)

    for _ in range(30):
        rover.send_speed()
        await trio.sleep(0.1)

    element_ids = [28, 30]
    if counts_supported:
        element_ids += [14, 16]

    for _ in range(10):
        rover.send_speed()
        await trio.sleep(0.1)

        data = await rover.get_data_items(element_ids)
        if counts_supported:
            enc_counts_left.append(data[14])
            enc_counts_right.append(data[16])

        enc_intervals_left.append(data[28])
        enc_intervals_right.append(data[30])

    if motor_effort == 0:
        assert all(i == 0 or i > 500 for i in enc_intervals_left)
        assert all(i == 0 or i > 500 for i in enc_intervals_right)
    else:
        assert 0.005 < (1 / statistics.mean(enc_intervals_left)) / abs(motor_effort) < 0.05
        assert 0.005 < (1 / statistics.mean(enc_intervals_right)) / abs(motor_effort) < 0.05

    if counts_supported:
        encoder_delta_left = [
            fix_encoder_delta(a - b) for a, b in zip(enc_counts_left[1:], enc_counts_left)
        ]
        encoder_delta_right = [
            fix_encoder_delta(a - b) for a, b in zip(enc_counts_right[1:], enc_counts_right)
        ]
        if motor_effort == 0:
            assert all(i == 0 for i in encoder_delta_left)
            assert all(i == 0 for i in encoder_delta_right)
        else:
            assert all(20 < i / motor_effort < 500 for i in encoder_delta_left)
            assert all(20 < i / motor_effort < 500 for i in encoder_delta_right)


@pytest.mark.parametrize(
    ("elt_a", "elt_b", "delta"),
    (
        pytest.param(52, 54, None, id="battery status"),
        pytest.param(34, 36, 0.03, id="state of charge"),
        pytest.param(60, 62, 3, id="temperature"),
        pytest.param(24, 26, 0.01, id="voltage (external)", marks=pytest.mark.xfail),
        pytest.param(64, 66, 0.01, id="voltage (internal)"),
        pytest.param(42, 44, 0.05, id="current (external)", marks=pytest.mark.xfail),
        pytest.param(68, 70, 0.05, id="current (internal)"),
    ),
)
async def test_batteries_similar(rover, elt_a, elt_b, delta):
    value_a = await rover.get_data(elt_a)
    value_b = await rover.get_data(elt_b)
    if delta is None:
        assert value_a == value_b
    else:
        assert value_a == pytest.approx(value_b, abs=delta)


SANE_RANGE = {
    0: (0, 2),
    6: (0, 1023),
    8: (0, 1023),
    10: (0, 1),
    12: (0, 1),
    14: (0, 2 ** 16 - 1),
    16: (0, 2 ** 16 - 1),
    20: (10, 70),
    22: (10, 70),
    24: (12, 16.5),
    26: (12, 16.5),
    28: (0, 2 ** 16 - 1),
    30: (0, 2 ** 16 - 1),
    34: (0, 1),
    36: (0, 1),
    42: (0, 1),
    44: (0, 1),
    60: (10, 55),
    62: (10, 55),
    64: (12, 16.5),
    66: (12, 16.5),
    68: (-16.5, +16.5),
    70: (-16.5, +16.5),
}


@pytest.mark.parametrize(
    ["elt"], [(k,) for k in SANE_RANGE.keys()], ids=lambda elt: OPENROVER_DATA_ELEMENTS[elt].name
)
async def test_sane_value(rover, elt):
    v = await rover.get_data(40)
    if not OPENROVER_DATA_ELEMENTS[elt].supported(v):
        pytest.xfail(f"Data element {elt} not implemented in firmware {v}")
    lo, hi = SANE_RANGE[elt]
    value = await rover.get_data(elt)
    assert lo <= value <= hi


@pytest.mark.parametrize("battery", ["A", "B"])
async def test_power_currents_charging(rover, battery):
    is_charging = await rover.get_data(38)
    assert is_charging in [True, False]

    v = await rover.get_data(40)

    if OPENROVER_DATA_ELEMENTS[68].supported(v) and OPENROVER_DATA_ELEMENTS[70].supported(v):
        battery_current_i2c = await rover.get_data({"A": 68, "B": 70}[battery])
        if is_charging:
            # charging battery has positive current
            assert battery_current_i2c >= 0
        else:
            # discharging battery should have negative current
            assert battery_current_i2c <= 0


@pytest.mark.xfail(reason="suspected hardware problems with analog measurements")
@pytest.mark.parametrize(
    ("elt_analog", "elt_i2c"), (pytest.param(24, 64, id="A"), pytest.param(26, 66, id="B")),
)
async def test_power_analog_i2c_voltages_agree(rover, elt_analog, elt_i2c):
    value_analog = await rover.get_data(elt_analog)
    value_i2c = await rover.get_data(elt_i2c)
    assert value_analog == pytest.approx(value_i2c, abs=0.05)


@pytest.mark.xfail(reason="suspected hardware problems with analog measurements")
@pytest.mark.parametrize(
    ("elt_analog", "elt_i2c"), (pytest.param(42, 68, id="A"), pytest.param(44, 70, id="B")),
)
async def test_power_analog_i2c_currents_agree(rover, elt_analog, elt_i2c):
    value_analog = await rover.get_data(elt_analog)
    value_i2c = await rover.get_data(elt_i2c)
    assert value_analog == pytest.approx(abs(value_i2c), abs=0.02)


async def test_fan_temperatures_equal(rover):
    fan_temp = await rover.get_data(20)
    fan_temp2 = await rover.get_data(22)
    assert fan_temp == pytest.approx(fan_temp2, abs=5)


motor_direction_param = [
    pytest.param(+1, id="forward", marks=pytest.mark.motor),
    pytest.param(-1, id="reverse", marks=pytest.mark.motor),
    pytest.param(0, id="still"),
]


@pytest.mark.parametrize("right", motor_direction_param)
@pytest.mark.parametrize("left", motor_direction_param)
async def test_motor_status(rover, left, right):
    v = await rover.get_data(40)
    if not all(OPENROVER_DATA_ELEMENTS[i].supported(v) for i in (72, 74)):
        pytest.xfail("Motor status flags not implemented in this version")
    rover.set_motor_speeds(left * 0.2, right * 0.2, 0)
    for i in range(3):
        rover.send_speed()
        await trio.sleep(0.1)

    l_status, r_status = (await rover.get_data(72), await rover.get_data(74))

    assert (left == 0) == (MotorStatusFlag.BRAKE in l_status)
    assert (right == 0) == (MotorStatusFlag.BRAKE in r_status)

    if left < 0:
        assert MotorStatusFlag.REVERSE in l_status
    if left > 0:
        assert MotorStatusFlag.REVERSE not in l_status

    if right > 0:
        assert MotorStatusFlag.REVERSE in r_status
    if right < 0:
        assert MotorStatusFlag.REVERSE not in r_status



import logging
from time import monotonic

import pytest
import trio

from openrover import open_rover

logger = logging.getLogger(__file__)


async def control_robot(rover, left, right, flipper, duration):
    rover.set_motor_speeds(left, right, flipper)
    logger.info(
        f"Sending command left motor: {left} right motor: {right} flipper: {flipper} duration"
        f" {duration} seconds..."
    )
    t1 = monotonic()
    while True:
        rover.send_speed()
        await trio.sleep(0.5)
        left_motor_encoder_interval = await rover.get_data(28)
        right_motor_encoder_interval = await rover.get_data(30)
        left_motor_temperature = await rover.get_data(20)
        right_motor_temperature = await rover.get_data(22)
        if t1 + duration < monotonic():
            break

    logger.info(
        f" left motor encoder interval: {left_motor_encoder_interval} right motor encoder"
        f" interval: {right_motor_encoder_interval} left motor temperature:"
        f" {left_motor_temperature} right motor temperature {right_motor_temperature}"
    )

    errs = []
    if left != 0 and left_motor_encoder_interval == 0:
        errs.append("left motor speed error")
    if right != 0 and right_motor_encoder_interval == 0:
        errs.append("right motor speed error")
    if left_motor_temperature > 200:
        errs.append("left motor temperature too high")
    if right_motor_temperature > 200:
        errs.append("right motor temperature too high")

    if errs:
        err = ";".join(errs)
        logger.error(err)
        raise RuntimeError(err)


@pytest.mark.burnin
async def test_burnin():
    t1 = monotonic()
    async with open_rover() as rover:
        logger.info("burn test procedure starting!")

        await trio.sleep(1)
        version = await rover.get_data(40)
        logger.info(f"rover firmware version {version}")

        # for testing, we could set x =0.02 or something less than 1 to speed up the process
        # x=0.02
        # for actual use of burn test script, the scale factor x should be set to 1
        x = 1

        # 1. Forward 35% speed for 30 sec
        # run 30 seconds left motor 350, right motor 350, fan 240
        await control_robot(rover, 0.35, 0.35, 0, 30 * x)
        # 2. Stop for 0.5 sec
        await control_robot(rover, 0, 0, 0, 0.5 * x)
        # 3. Left turn for 2 sec
        await control_robot(rover, -0.3, 0.3, 0, 2 * x)
        # 4. Stop for 0.5 sec
        await control_robot(rover, 0, 0, 0, 0.5 * x)
        # 5. Right turn for 2 sec
        await control_robot(rover, 0.3, -0.3, 0, 2 * x)
        # 6. Stop for 0.5 sec
        await control_robot(rover, 0, 0, 0, 0.5 * x)
        # 7. Forward 35% speed for 30 sec
        await control_robot(rover, 0.35, 0.35, 0, 30 * x)
        # 8. Flipper up for 6 sec
        await control_robot(rover, 0, 0, 0.35, 6 * x)
        # 9. Flipper down for 6 sec
        await control_robot(rover, 0, 0, -0.35, 6 * x)
        # 10. Flipper up for 6 sec
        await control_robot(rover, 0, 0, 0.35, 6 * x)
        # 11. Flipper down for 6 sec
        await control_robot(rover, 0, 0, -0.35, 6 * x)
        # 12. Flipper up for 1.5 sec
        await control_robot(rover, 0, 0, 0.35, 1.5 * x)
        # 13. Forward 35% speed for 60 sec
        await control_robot(rover, 0.35, 0.35, 0, 60 * x)
        # 14. Stop for 0.5 sec
        await control_robot(rover, 0, 0, 0, 0.5 * x)
        # 15. Backward 35% speed for 30 sec
        await control_robot(rover, -0.35, -0.35, 0, 30 * x)
        # 16. Stop for 10 sec
        await control_robot(rover, 0, 0, 0, 10 * x)
        # another loop starts
        # 15 times loop, full speed
        for n in range(0, 15):
            # 1. Forward 100% speed for 30 sec
            await control_robot(rover, 1, 1, 0, 30 * x)
            # 2. Stop for 0.5 sec
            await control_robot(rover, 0, 0, 0, 0.5 * x)
            # 3. Left turn for 2 sec
            await control_robot(rover, -0.3, 0.3, 0, 2 * x)
            # 4. Stop for 0.5 sec
            await control_robot(rover, 0, 0, 0, 0.5 * x)
            # 5. Right turn for 2 sec
            await control_robot(rover, 0.3, -0.3, 0, 2 * x)
            # 6. Stop for 0.5 sec
            await control_robot(rover, 0, 0, 0, 0.5 * x)
            # 7. Forward 60% speed for 30 sec
            await control_robot(rover, 0.6, 0.6, 0, 30 * x)
            # 8. Flipper up for 6 sec
            await control_robot(rover, 0, 0, 0.8, 6 * x)
            # 9. Flipper down for 6 sec
            await control_robot(rover, 0, 0, -0.8, 6 * x)
            # 10. Flipper up for 6 sec
            await control_robot(rover, 0, 0, 0.8, 6 * x)
            # 11. Flipper down for 6 sec
            await control_robot(rover, 0, 0, -0.8, 6 * x)
            # 12. Flipper up for 1.5 sec
            await control_robot(rover, 0, 0, 0.8, 1.5 * x)
            # 13. Forward 60% speed for 60 sec
            await control_robot(rover, 0.6, 0.6, 0, 60 * x)
            # 14. Stop for 0.5 sec
            await control_robot(rover, 0, 0, 0, 0.5 * x)
            # 15. Backward 60% speed for 30 sec
            await control_robot(rover, 0.6, 0.6, 0, 30 * x)
            # 16. Stop for 10 sec
            await control_robot(rover, 0, 0, 0, 10 * x)
            # loop ends
        t2 = monotonic()
        logging.info(f"finished in {t2 - t1} seconds")
        # close the file if needed
        # do something to indicate the burn test is finished
        logging.info("burn test finished successfully!")
